<?php
	// Random instance generator for the HcVIP problem
	// Instance parameters	
	$DEST = NULL;
	$VIP = array();
	$HOME = NULL;
	$BUS = array();  // 2D matrix: from/to

	// Generator parameters are $DEST, $VIP, and nb. of bus lines
	$BUSLINES = NULL;
	$NBVIP = NULL;
	
	for ($i=1; $i<count($argv); $i++) {
		if ($argv[$i] === "-destinations") {
			$DEST = $argv[++$i];
			if (!is_numeric($DEST) || $DEST < 1) 
				die("$DEST is not a good value for `destinations'\n");
		} else if ($argv[$i] === "-vipdensity") {
			$VIPDENSITY = $argv[++$i];
			if ($VIPDENSITY < 0 || $VIPDENSITY > 100) die("Error: VIP density value $VIPDENSITY is not in range 0..100\n");			
			$NBVIP = $VIPDENSITY*1.0/100 * $DEST;
		} else if ($argv[$i] === "-buslinedensity") {
			$BUSDENSITY = $argv[++$i];
			if ($BUSDENSITY < 0 || $BUSDENSITY > 100) die("Error: value for bus line density $DENSITY is not in range 0..100\n");			
		
			$BUSLINES = ($BUSDENSITY*1.0/100) * $DEST*($DEST-1);
		}
	}
	
	// Check that all params have been given
	$DEST != NULL or die("Number of destinations missing. Use -destinations <positive integer>\n");
	$NBVIP != NULL or die("Number of VIPs missing. Use -vip <integers in 0..nb of destinations>\n");
	$BUSLINES != NULL or die("Density of bus lines missing. Use -buslinedensity <float in 0..100>\n");

	
	// Build BUS 0/1 matrix (from/to)
	for($from=1; $from<=$DEST; $from++) {
		for($to=1; $to<=$DEST; $to++) {
			$BUS[$from][$to] = 0;
		}
	}
	
	// Adding $BUSLINES bus lines randomly (very inefficient when nb. of lines is high!)
	$howmany = 0;
	while($howmany < $BUSLINES) {
		$from = rand(1, $DEST);
		$to = rand(1, $DEST);
		
		if ($from == $to) continue; // skip
		if ($BUS[$from][$to] == 1) continue; // skip already existing lines
		
		$BUS[$from][$to] = 1;
		$howmany++;
	}
	
	// Choosing home = 1
	$HOME = 1;
	
	// Choosing NBVIP VIP destinations
	$howmany = 0;
	while($howmany < $NBVIP) {
		$v = rand(2, $DEST);  // 2 --> excluding home
		if (in_array($v, $VIP)) continue;
		
		$VIP[] = $v;
		$howmany++;
	}
	
	// Output instance
	echo("# Randomly generated instance.\n");
	echo("\n\ndestinations = $DEST\n");
	
	echo("# $NBVIP vip destinations ($VIPDENSITY% density)\n");
	echo("vip = " . join(", ", $VIP)."\n\n");
	echo("# 'home' is destination number:\n");
	echo("home = $HOME\n");
	echo("\n\n# Available bus lines ($BUSLINES, density $BUSDENSITY%): from->to\n");
	
	for($from=1; $from<=$DEST; $from++) {
		for($to=1; $to<=$DEST; $to++) {
			if ($BUS[$from][$to] == 1) {
				echo("bus = $from -> $to\n");
			}
		}
	}


?>
