c Rostering to SAT compiler, the new version is in EDIMACS format (non-cnf)
c Author: Daniele Zannella - Sapienza University of Rome
c Date: July 18, 2008
c
<?php

	$DEBUG = false;  // <-- Set this variable to true to generate a more intelligible (non-DIMACS) 
					 //     CNF (variables with symbolic names and few comments)


	// Variables
	$I = NULL;
	$T = NULL;
	$N = NULL;
	$C = NULL;
	$R = NULL;
	$L = NULL;


    $fileName = $argv[1];
	if ($fileName == "") {
		echo("c Syntax: php ".$argv[0]." <instance>\n");
		exit;
	};
	readInstanceFile($fileName);


	// Check that everything has been given
	$I > 0 or die("Number of employees missing\n");
	$T > 0 or die("Number of slots in each day missing\n");
	$N > 0 or die("Number of night slots missing\n");
	$C > 0 or die("Min number of employees that must be at work in each slot missing\n");
	$R > 0 or die("Min number of `break' slots missing\n");	
	$L > 0 or die("Min number of working slots missing\n");	
	
	// Everything is ok.
	
	
	$CLAUSES = array();
	$NCLAUSES = 0;
	$NVARS = 0;
	$dictionary = NULL; 
	/* will be a 4D array: 
		dictionary['WORKS'][empl][slot] =  dimacsVar for works(empl,slot)
		dictionary['WORKS_IN'][empl][tupleOfSlots] = dimacsVar for worksIn(e, <s1..sL>)
1¡Vin.->dictionary['ALL_WORK_IN][slot][tupleOfEmployees] = dimacsVar for allWorkIn(s, <e1...eC>)
		dictionary['NEVER_WORKS'][employee] = dimacsVar for neverWorks(empl)
		
		The last three set of variables are needed only for CNF translation (cf. comments below).
	*/
	$outGateArray = array();
?>
c
c Problem description (a variation of the Rostering problem):
c In a factory that runs 24 hours a day and 7 days per week, each weekday (Monday to Sunday) is divided into 
c T time slots.
c The first N minore T slots of each day actually occur in the night, and need special treatment.
c The problem amounts to assign each of the I employees of the factory to slots in the weekly schedule, 
c such that some constraints are satisfied.
c
c  Constraints are as follows:
c  1. At least C employees must work in every slot of each day
c  2. Each employee, after working in a slot (not in the night) must have a break of (i.e., (s)he doesn't work)
c     for at least the R next slots.
c  3. Each employee, after working in a slot (in the night) must have a break up to the end of the next 
c     working day.
c  4. Each employee must be assigned to at least L slots during the week. Otherwise, (s)he must not work at all
c     (i.e., we don't want to pay an employee for less than L slots a week).
c
c
c 
<?php
	echo("c Number of employees: $I\n");
	echo("c Number of slots in each day: $T\n");
	echo("c Number of night slots: $N\n");
	echo("c Min number of employees that must be at work in each slot: $C\n");	
	echo("c Min number of break slots: $R\n");	
	echo("c Min number of working slots: $L\n");	
	
	echo("c \n");
	echo("c \n");
	echo("c START DICTIONARY\n");

	if (!$DEBUG) {
		for ($e=1; $e<=$I; $e++) {
			for ($s=1; $s<=$T*7; $s++) {		
					echo("c  WORKS($e,$s) = ".works2dimacs($e, $s)."\n");
				}
			}
		}
	 else echo("c omitted -- DEBUG mode\n");


	echo("c END DICTIONARY\n"); 
	echo("c \n");
	echo("c \n");

	/* Variables:
	    A first choice could be:
	     works(emp,day,slot)  for each emp in I, day in 1..7, slot in 1..T
	        var = true means employee `emp' works in slot `slot' of day `day'

		However, given that some constraints (in particular those imposing a fixed number R of non-working slots after a working one) may involve "spanning" over different days,
		a simpler approach would be that of considering the following variables:

			work(emp, slot)  for each empy in I and slot in 1..7*T.
			
		In other words, we consider all the slots in a week, and number them from 1 to 7*T.
		Slots having a number multiple of T will be the last ones in a day (slots T, 2T, 3T,...,7T).
	
	*/



	if ($DEBUG) addComment("c (1) In each slot at least $C employees work");
	/* AND_{s in 1..7T}
	 		OR_{ <e1...eC> (ordered) s.t. e1..eC in 1..I and all different } 
	 			works(e1,s) & ... & works(eC,s).
	   In other words, for every slot there must be at least one set of C employees (represented as an ordered C-tuple, in order to avoid "duplicates") that 
	   work in that slot.

	   
	   This constraint is not in CNF (it is a disjunction of conjuncts). It must be converted...
	   
	*/
	
		if ($DEBUG) addComment("   c (1.1)");		

		// ------------------------------------- NEW ------------------------------------------------	
		// Vincolo (1) In Ogni turno della settimana ci sono almeni C infermieri
		// ristrutturiamo il tutto cercando di minimizzare il numero delle variabili e mantenere la struttura della formula

		$auxArray = array();  // questo nuovo array serve per memorizzare volta per volta le tuple di C infermieri
		$count = 0;
		for($s=1; $s<=7*$T; $s++) {
		//	$NVARS++;
		//	array_push($outGateArray, $NVARS);
		//	$clause = "6 -1 $NVARS ";
			$tuple = firstOrderedTuple($C,1, $I);  // a tuple with C elements in 1..I (the first is tuple: 1..C)
			while ($tuple !== NULL) {
				$NVARS++;
				array_push($auxArray, $NVARS);
				$auxClause = "4 -1 $NVARS ";
				for($i=0;$i<count($tuple);$i++) {
					$auxClause .= works2dimacs($tuple[$i], $s)." ";
				}
				addClause($auxClause);
				$tuple = nextOrderedTuple($tuple, $C,1,$I);
			}
			$NVARS++;
			array_push($outGateArray, $NVARS);
			$clause = "6 -1 $NVARS ";
			for($j=0;$j<count($auxArray);$j++)
			{
				$clause .= $auxArray[$j]. " ";			
			}

			addClause($clause); // formula finale per ogni turno della settimana
			$auxArray = array_slice($auxArray, count($auxArray)); // Serve per resettare l'array ausiliario in modo da poterlo 										      // riutizzare per il turno successivo
			$count++;
		}
		// ---------------------------------------- NEW ------------------------------------------------
		
	/*	$clause = "";
		$NVARS++;
		$output1 = $NVARS; 
		$clause .= "4 -1 $NVARS ";
		
		for($i=0;$i<=count($outGateArray);$i++)
		{	
			$clause .= $outGateArray[$i]." ";
		}
		addClause($clause);
	*/	
		

	if ($DEBUG) addComment("c (2) Every employee, after working (not in the night), must not work for the next $R slots");
	/* 	AND_{e in I, s in 1..7*T s.t. !(1 <= s%T <= N) [i.e., in a slot not in the night, cf. below] }
			works(e,s) -> 
				AND_{ s' in s+1..s+$R [sum made modulo 7T] }   -works(e,s').
				
		In other words, if an employee `e' works in slot `s' (not in the night), she must be free in the following $R slots. Since the scheduling is periodic, we have to consider summation modulo 7T (night slots are 1..N (first day), T+1..T+N (second day), ..., kT+1..kT+N (k-th day),...,  hence those slots `s' such that 
		1<= s%T <= N).
				
	*/
	
	// cos facendo posso applicare direttamente l'implicazione
 /* 	for($e=1; $e<=$I; $e++) {	
	   	for($s=1; $s<=7*$T; $s++) {	
	   		$clause .= "10 -1 $OUTGATE "
  			if ( $s % $T < 1   ||  $s % $T > $N ) {
  				$clause .= "-".works2dimacs($e,$s)." ";
			   	for($ss=$s+1; $ss<=$s+$R; $ss++) {		  			
			   		$sss = $ss;
			   		$clause1 = "4 -1 $OUTGATE "
			   		if ($sss > 7*$T) $sss = $sss % 7*$T;
			   			$clause1 .= "-".works2dimacs($e,$sss)." ";
  				}
  				addClause($clause1);
  				$clause .= $OUTGATE;
  			}
  			   
  		}
	}
*/	
	// ---------------------------------------- NEW ------------------------------------------------
	// il problema della codifica di cui sopra sta nel fatto che il tool SATMATE accetta si formule in formato Edimacs ma solo con 		// operatori di tipo AND, OR e NOT. Inoltre anche per il formato ISCAS abbiamo questo tipo di problema. 
	for($e=1; $e<=$I; $e++) {	
	   	for($s=1; $s<=7*$T; $s++) {	
  			if ( $s % $T < 1   ||  $s % $T > $N ) { // questo turno è diurno???
  			//	$NVARS++;
  			//	array_push($outGateArray, $NVARS); 
  			//	$clause = "";
  			//	$clause .= "6 -1 $NVARS -".works2dimacs($e,$s)." ";
				$NVARS++;
				$output = $NVARS;
				$auxClause = "4 -1 $NVARS ";
			   	for($ss=$s+1; $ss<=$s+$R; $ss++) {
			   		$sss = $ss;
			   		if ($sss > 7*$T) $sss = $sss % 7*$T;
			   		$auxClause .= "-" . works2dimacs($e,$sss)." ";
  				}

				addClause($auxClause);
				
				$NVARS++;
  				array_push($outGateArray, $NVARS); 
  				$clause = "";
  				$clause .= "6 -1 $NVARS -".works2dimacs($e,$s)." ";
				$clause .= $output;
  				addClause($clause);   
  			}
  		}
	}

	/*	$clause = "";
		$NVARS++;
		$output2 = $NVARS;
		$clause .= "4 -1 $NVARS ";
		
		for($j=$i-1;$j<=count($outGateArray);$j++)
		{	
			$clause .= $outGateArray[$j]." ";
		}
		addClause($clause);	
	*/

	if ($DEBUG) addComment("c (3) Every employee, after working during night, must not work until the end of the next day");
	/* AND_{e in I, s in 1..7T s.t. (1 <= s%T <= N) [i.e., in a slot in the night] }
			works(e, s ) -> 
				AND_{ s' in s+1..s+k, with k>0 the min. integer s.t. (s+k)%T = 0 and k>=T [end of next day] [sum made modulo 7T] }
			 -works(e,s')

	*/
	
		// ---------------------------------------- NEW ------------------------------------------------
	
  	for($e=1; $e<=$I; $e++) {	
	   	for($s=1; $s<=7*$T; $s++) {
  			if ( ($s % $T) >= 1   &&    ($s % $T) <= $N ) { // il turno è notturno???
  				//... Spostato
  				$NVARS++;
				$output1 = $NVARS;
				$auxClause = "4 -1 $NVARS ";
  				$k=1;
				while ($k<=$T  || ( (($s+$k) % $T) != 1 ) ) {
						$ss = $s+$k;				
						if ($ss > 7*$T)  $ss = $ss - 7*$T;
						$auxClause .= "-" .works2dimacs($e,$ss)." ";
						$k++;
				} 

				addClause($auxClause);
				$NVARS++;
  				array_push($outGateArray,  $NVARS); 
  				$clause = "";
  				$clause .= "6 -1 $NVARS -".works2dimacs($e,$s)." ";
				$clause .= $output1;
				addClause($clause);
			}
		}
	}

	/*	$clause = "";
		$NVARS++;
		$output3 =$NVARS;
		$clause .= "4 -1 $NVARS ";
		
		for($z=$j-1;$z<=count($outGateArray);$z++)
		{	
			$clause .= $outGateArray[$z]." ";
		}
		addClause($clause);
	*/	

	if ($DEBUG) addComment("c (4) Employee either don't work at all, or work for at least $L slots");
	
	

	if ($DEBUG) addComment("\n\nc (4.1)");
	/*	(4.1) Vincolo NON-Cnf 
			AND_{e in I}
			  neverWorks(e)
			OR
			  OR_{ <s1,...sL> (ordered) s.t. s1..sL in 1..7*T and all different } 
					worksIn(e, <s1..sL>)
	*/

	// riscrivo il vincolo senza agiunta di nuove variabili e ne conservo struttura

	$auxArray2 = array();
 	for($e=1; $e<=$I; $e++) {
 	//	$NVARS++;
 	//	array_push($outGateArray, $NVARS); 
 	//	$clause = "6 -1 $NVARS ". neverWorks2dimacs($e)." ";	
  		
	   	$tuple = firstOrderedTuple($L,1, 7*$T);  // a L-tuple with elements in 1..7*T
	   	while ($tuple !== NULL) {
			if($L>1) {
				$NVARS++;			
				array_push($auxArray2, $NVARS); 
				$auxClause2 = "4 -1 $NVARS ";	
				for($i=0;$i<count($tuple);$i++) {
					$auxClause2 .= works2dimacs($e,$tuple[$i])." ";
				}			
			}
			else if($L == 1) { // caso in cui un impiegato basta che copri un solo turno  
					array_push($auxArray2, works2dimacs($e,$tuple[0])); 
			}
			addClause($auxClause2);
	   		
			$tuple = nextOrderedTuple($tuple, $L,1, 7*$T);	   		
	   	}
		
		//$clause .= $auxGate." ";
		$NVARS++;
 		array_push($outGateArray, $NVARS); 
 		$clause = "6 -1 $NVARS ". neverWorks2dimacs($e)." ";	

		for($j=0;$j<count($auxArray2);$j++) {
			$clause .= $auxArray2[$j]." ";
		}

		addClause($clause);
		$auxArray2 = array_slice($auxArray2, count($auxArray2));
	}
	
	// Print out dictionary
//	echoDictionary();
	
	// Print out a list of structural properties of variables (useful for ad-hoc solvers)
	// echoProperties();
	$NVARS++;
	$finalClause = "4 -1 $NVARS ";
	
	for($h=0;$h<count($outGateArray);$h++)
		{	
			$finalClause .= $outGateArray[$h]." ";
		}	
	
	addClause($finalClause);
	
	// Print out header line "p NON-cnf..."
	isset($NVARS) or die("maxVar is not set");
	echo("p noncnf $NVARS \n");
	
	// Print out all clauses
	echoClauses();	

	//print_r($dictionary);
	
	


	function addClause($clause) {
		global $CLAUSES,$NCLAUSES;
		$clause = trim($clause);
		if ($clause == "") return;
		$clause .= " 0";
		$CLAUSES[] = $clause; // Append to array
	//	$NCLAUSES++;
	}

	function addComment($msg) {
		global $CLAUSES;
		$CLAUSES[] = $msg;
		// No $NCLAUSES increment here!
	}


	// 	dictionary['WORKS'][empl][slot] =  (works(empl,slot)==true)
	function works2dimacs($e,$s) {
		global $NVARS, $dictionary, $DEBUG, $I, $T;
		
		if ($e < 1 || $e > $I) {
			echoClauses();
			die("Trying to retrieve a 'WORKS' variable for employee $e: WORKS($e,$s)");
		}
		if ($s < 1 || $s > 7*$T) {
			echoClauses();		
			die("Trying to retrieve a 'WORKS' variable for slot $s: WORKS($e,$s)");
		}
		$result = "WORKS(e=$e,s=$s)";		
		
		$v = $dictionary['WORKS'][$e][$s];
		if ($v !== NULL) {
			if (!$DEBUG) $result = $v;
		}
		else {
			$NVARS++;
			$dictionary['WORKS'][$e][$s] = $NVARS;
			if (!$DEBUG) $result = $NVARS;
		}
		return $result;
	}
	
	
	//dictionary['ALL_WORK_IN][slot][tupleOfEmployees] = (allWorkIn(s, <e1...eC>)==true)
	function allWorkIn2dimacs($s,$tuple) {
		global $NVARS, $dictionary, $DEBUG, $I, $T;
		$tupleString = join(",",$tuple); // tuple of employees		
		if ($s < 1 || $s > 7*$T) {
			echoClauses();		
			die("Trying to retrieve a 'ALL_WORK_IN' variable for slot $s: ALL_WORK_IN($s,emp=<$tupleString>)");
		}

		$result = "ALL_WORK_IN(s=$s,empl=<".$tupleString.">)";
		
		$v = $dictionary['ALL_WORK_IN'][$s][$tupleString];
		if ($v != NULL) { // se la tupla è già presente nel dizionario e dunque ha già assegnato un valore
			if (!$DEBUG) $result = $v;
		}
		else {  // altrimenti gli assegno uno valore (numero)
			$NVARS++;
			$dictionary['ALL_WORK_IN'][$s][$tupleString] = $NVARS;
			if (!$DEBUG) $result = $NVARS;
		}
		return $result;
	}	
	
	
	// dictionary['WORKS_IN'][empl][tupleOfSlots] = (worksIn(e, <s1..sL>) = true)
	function worksIn2dimacs($e,$tuple) {
		global $NVARS, $dictionary, $DEBUG, $I, $T;
		
		if (!is_array($tuple)) {
			die("tuple is not array: " . ($tuple===null?"null":$tuple)."\n");
		}
		
		$tupleString = join(",", $tuple); // tuple of slots		
		if ($e < 1 || $e > $I) {
			echoClauses();		
			die("Trying to retrieve a 'WORKS_IN' variable for employee $e: WORKS_IN(emp=$e,slots=<$tupleString>)");
		}

		$result = "WORKS_IN(e=$e,slots=<".$tupleString.">)";
		
		$v = $dictionary['WORKS_IN'][$e][$tupleString];
		if ($v != NULL) {
			if (!$DEBUG) $result = $v;
		}
		else {
			$NVARS++;
			$dictionary['WORKS_IN'][$e][$tupleString] = $NVARS;
			if (!$DEBUG) $result = $NVARS;
		}
		return $result;
	}

	
	
	// dictionary['NEVER_WORKS][employee] = (neverWorks(empl) == true)
	function neverWorks2dimacs($e) {
		global $NVARS, $dictionary, $OUTGATE, $I, $T;
		$NVARS++;
		$res = $NVARS;
		$result = "4 -1 $NVARS ";	
		
		if ($e < 1 || $e > $I) {
			echoClauses();		
			die("Trying to retrieve a 'NEVER_WORKS' variable for employee $e: NEVER_WORKS($e)");
		}
			
	   	for($s=1; $s<=7*$T; $s++)
	   		$result	.= "-".works2dimacs($e,$s)." ";	
	
		addClause($result);
		return $res;
	}
	
	
	function echoClauses() {
		global $CLAUSES;
		foreach($CLAUSES as $c) {
			echo($c ."\n");
		}
	}	
	
	
	function echoDictionary() {
		global $NVARS, $dictionary, $DEBUG, $I, $T;
		echo ("c\nc START DICTIONARY\n");
		/*if ($DEBUG) {
			echo("c omitted -- DEBUG mode\n");
		}
		else {*/		
			if (isset($dictionary['WORKS'])) {			
				// 	dictionary['WORKS'][empl][slot] =  dimacsVar
				foreach($dictionary['WORKS'] as $e => $entryE) {
					foreach($entryE as $s => $dimacsVar) {
						echo("c works($e,$s) = $dimacsVar\n");
					}
				}
			}		
			if (isset($dictionary['WORKS_IN'])) {
				// dictionary['WORKS_IN'][empl][tupleOfSlots] = dimacsVar
				foreach($dictionary['WORKS_IN'] as $e => $entryE) {
					foreach($entryE as $tuple => $dimacsVar) {
						echo("c works_in($e, <$tuple>) = $dimacsVar\n");
					}
				}	
			}
			if (isset($dictionary['ALL_WORK_IN'])) {
				//dictionary['ALL_WORK_IN][slot][tupleOfEmployees] = dimacsVar			
				foreach($dictionary['ALL_WORK_IN'] as $s => $entryS) {
					foreach($entryS as $tuple => $dimacsVar) {
						echo("c all_work_in($s, <$tuple>) = $dimacsVar\n");
					}
				}	
			}
			if (isset($dictionary['NEVER_WORKS'])) {
				// dictionary['NEVER_WORKS][employee] = dimacsVar
				foreach($dictionary['NEVER_WORKS'] as $e => $dimacsVar) {
					echo("c never_works($e) = $dimacsVar\n");
				}	
			}			
		//}
		echo("c END DICTIONARY\nc\nc\n");		
	}	
	
	
	function echoProperties() {
		global $NVARS, $dictionary, $DEBUG, $I, $T;
		echo("c START PROPERTIES\n");
		if (isset($dictionary['WORKS_IN'])) {
			/* WORKS_IN variables are dependent on some WORK variables
				worksIn(e, <s1,..,sL>) <-> works(e, s1) & ... & works(e, sL)
		    */	
			foreach($dictionary['WORKS_IN'] as $e => $entryE) {
				// dictionary['WORKS_IN'][empl][tupleOfSlots] = dimacsVar
				foreach($entryE as $tuple => $dimacsVar) {
					$tupleArray = explode(",", $tuple);
					echo("c " . worksIn2dimacs($e, $tupleArray)." DEPENDENT ON ");
					foreach($tupleArray as $s) {
						echo( works2dimacs($e, $s) . " ");
					}
					echo("\n");
				}
			}
		}
				

		if (isset($dictionary['ALL_WORK_IN'])) {
			/* ALL_WORK_IN variables are dependent on some WORKS variables
	        	allWorkIn(s, <e1...eC>)  <-> works(e1,s) & ... & works(eC,s)
		    */	
			foreach($dictionary['ALL_WORK_IN'] as $s => $entryS) {
				//dictionary['ALL_WORK_IN][slot][tupleOfEmployees] = dimacsVar			
				foreach($entryS as $tuple => $dimacsVar) {
					$tupleArray = explode(",", $tuple);
					echo("c " . allWorkIn2dimacs($s, $tupleArray)." DEPENDENT ON ");
					foreach($tupleArray as $e) {
						echo( works2dimacs($e, $s) . " ");
					}
					echo("\n");
				}
			}
		}

		if (isset($dictionary['NEVER_WORKS'])) {
			/* NEVER_WORKS variables are dependent on some WORKS variables
	        	neverWorks(e) <-> ( AND_{ s in 1..7T } -works(e,s) )
		    */	
			foreach($dictionary['NEVER_WORKS'] as $e => $dimacsVar) {
				// dictionary['NEVER_WORKS][employee] = dimacsVar
				echo("c " . neverWorks2dimacs($e) . " DEPENDENT ON ");
				for($s=1; $s <= 7*$T; $s++) {
					echo( works2dimacs($e,$s) . " ");
				}
				echo("\n");
			}
		}
		echo("c END PROPERTIES\n");
	}	
	
	
	

	function firstOrderedTuple($len,$min, $max) {  // a $len-tuple with elements in 1..$max all different
		$result = array();
		for ($i = 0; $i<$len; $i++) {
			$result[$i] = $min+$i;
		}
		return $result;
	}

	// questo metodo crea passo passo tutte le combinazioni di $C infermieri per ogni turno
	function nextOrderedTuple($tuple, $len,$min, $max) {  // next $len-tuple (wrt $tuple) with elements in 1..$max all different
		$result = $tuple;

		for ($i=$len-1; $i>=0; $i--) {		
			$remaining = $len-1 - $i;
			if ($result[$i] < $max && ($result[$i]+1 <= $max-$remaining )   ) {
				$result[$i]++;			
				for ($j = $i+1; $j < $len; $j++) {
					$result[$j] = $result[$i] + ($j - $i);
				}			
				return $result;				
			}
		}
		return NULL;
	}
	
	
	
	
	function readInstanceFile($fileName) {
		global $I, $T, $N, $C, $R, $L;	
		$COMMENTS = array("#", "//");		
		$file = file($fileName) or die("Could not open instance file $fileName\n");
		foreach($file as $line) {
			//echo("c Reading line: $line\n");
			$commentSign = NULL;
			foreach($COMMENTS as $c) {
				 $pos = strpos($line, $c);
				 if ($pos === FALSE) continue;
				 if ($commentSign === NULL || $commentSign > $pos) $commentSign = $pos;
			}
			if ($commentSign !== NULL) {
				$line = substr($line, 0, $commentSign);
				//echo("   c Removing comments (from col. $commentSign onwards): $line\n");		
			}
			$line = trim($line);
			if ($line == "") continue;
	
			list($what,$content) = split("=",$line);
			$what = trim($what);
			$content = trim($content);
	
			//echo("   c What: $what -- Content: $content\n");		
			
			if ($what == "employees" && $I===NULL) {
				$I = $content;
			}
			else if ($what == "dailySlots" && $T===NULL) {
				$T = $content;
			}
			else if ($what == "nightSlots" && $N===NULL) {
				$N = $content;
			}
			else if ($what == "minEmployeesAtWork" && $C===NULL) {
				$C = $content;
			}
			else if ($what == "minBreak" && $R===NULL) {
				$R = $content;
			}
			else if ($what == "minWorkingSlots" && $L===NULL) {
				$L = $content;
			}
		}
	}	
?>
