c Uso: 
c - Eseguire: php rostering.php <file istanza>
c
c Esecuzione con zChaff su istanza di test e successiva traduzione del modello trovato in forma intellegibile
c php shopScheduling.php rostering-sampleInstance.txt > cnf.txt ; zchaff cnf.txt | php zchaff2Rostering.php cnf.txt
c
c Rostering to SAT compiler (the most straightforward and inefficient encoding, just a simple exercise!)
c Author: Toni Mancini - Sapienza University of Rome
c Date: July 18, 2008
c
<?php

	$DEBUG = false;  // <-- Set this variable to true to generate a more intelligible (non-DIMACS) 
					 //     CNF (variables with symbolic names and few comments)


	// Variables
	$I = NULL;
	$T = NULL;
	$N = NULL;
	$C = NULL;
	$R = NULL;
	$L = NULL;


    $fileName = $argv[1];
	if ($fileName == "") {
		echo("c Syntax: php ".$argv[0]." <instance>\n");
		exit;
	};
	readInstanceFile($fileName);


	// Check that everything has been given
	$I > 0 or die("Number of employees missing\n");
	$T > 0 or die("Number of slots in each day missing\n");
	$N > 0 or die("Number of night slots missing\n");
	$C > 0 or die("Min number of employees that must be at work in each slot missing\n");
	$R > 0 or die("Min number of `break' slots missing\n");	
	$L > 0 or die("Min number of working slots missing\n");	
	
	// Everything is ok.
	
	
	$CLAUSES = array();
	$NCLAUSES = 0;
	$NVARS = 0;
	$dictionary = NULL; 
	/* will be a 4D array: 
		dictionary['WORKS'][empl][slot] =  dimacsVar for works(empl,slot)
		dictionary['WORKS_IN'][empl][tupleOfSlots] = dimacsVar for worksIn(e, <s1..sL>)
		dictionary['ALL_WORK_IN][slot][tupleOfEmployees] = dimacsVar for allWorkIn(s, <e1...eC>)
		dictionary['NEVER_WORKS'][employee] = dimacsVar for neverWorks(empl)
		
		The last three set of variables are needed only for CNF translation (cf. comments below).
	*/
	
?>
c
c Problem description (a variation of the Rostering problem):
c In a factory that runs 24 hours a day and 7 days per week, each weekday (Monday to Sunday) is divided into 
c T time slots.
c The first N < T slots of each day actually occur in the night, and need special treatment.
c The problem amounts to assign each of the I employees of the factory to slots in the weekly schedule, 
c such that some constraints are satisfied.
c
c Constraints are as follows:
c  1. At least C employees must work in every slot of each day
c  2. Each employee, after working in a slot (not in the night) must have a break of (i.e., (s)he doesn't work)
c     for at least the R next slots.
c. 3. Each employee, after working in a slot (in the night) must have a break up to the end of the next 
c     working day.
c. 4. Each employee must be assigned to at least L slots during the week. Otherwise, (s)he must not work at all
c     (i.e., we don't want to pay an employee for less than L slots a week).
c
<?php
	echo("c Number of employees: $I\n");
	echo("c Number of slots in each day: $T\n");
	echo("c Number of night slots: $N\n");
	echo("c Min number of employees that must be at work in each slot: $C\n");	
	echo("c Min number of break slots: $R\n");	
	echo("c Min number of working slots: $L\n");	
	


	/* Variables:
	    A first choice could be:
	     works(emp,day,slot)  for each emp in I, day in 1..7, slot in 1..T
	        var = true means employee `emp' works in slot `slot' of day `day'

		However, given that some constraints (in particular those imposing a fixed number R of non-working slots after a working one) may involve "spanning" over different days,
		a simpler approach would be that of considering the following variables:

			work(emp, slot)  for each empy in I and slot in 1..7*T.
			
		In other words, we consider all the slots in a week, and number them from 1 to 7*T.
		Slots having a number multiple of T will be the last ones in a day (slots T, 2T, 3T,...,7T).
	
	*/



	if ($DEBUG) addComment("c (1) In each slot at least $C employees work");
	/* AND_{s in 1..7T}
	 		OR_{ <e1...eC> (ordered) s.t. e1..eC in 1..I and all different } 
	 			works(e1,s) & ... & works(eC,s).
	   In other words, for every slot there must be at least one set of C employees (represented as an ordered C-tuple, in order to avoid "duplicates") that 
	   work in that slot.

	   
	   This constraint is not in CNF (it is a disjunction of conjuncts). It must be converted:
	   
	   as usual, we introduce additional (functional dependent) variables:
	        allWorkIn(s, <e1...eC>)  <-> works(e1,s) & ... & works(eC,s)
	   for any possible slot `s' and any possible ordered C-tuple <e1...eC> of employees.
	   
	   Constraint becomes:
	   (1.1) AND_{s in 1..7T} 
	 		OR_{ <e1...eC> (ordered) s.t. e1..eC in 1..I and all different } 
	   			allWorkIn(s, <e1..eC>)
	   AND
	   (1.2) AND_{s in 1..7T,  <e1..eC> as above} 
	   			allWorkIn(s, <e1..eC>) <-> works(e1,s) & ... & works(eC,s).
	   
	   Constraint (1) is now in CNF, while (2) can be rewritten as:
	   (1.2a) AND_{s in 1..7T,  <e1..eC> as above} 
	   			allWorkIn(s, <e1..eC>) -> works(e1,s) & ... & works(eC,s)
	   AND
	   (1.2b) AND_{s in 1..7T,  <e1..eC> as above} 
	   			works(e1,s) & ... & works(eC,s) -> allWorkIn(s, <e1..eC>)
	   
	   hence as:	   
	   
	   (1.2a) AND_{s in 1..7T,  <e1..eC> as above} 
	   			-allWorkIn(s, <e1..eC>) || works(e1,s) 
	   			AND
	   			...
	   			AND
	   			-allWorkIn(s, <e1..eC>) || works(eC,s) 
 	    AND
	    (1.2b) AND_{s in 1..7T,  <e1..eC> as above} 
	    		-works(e1,s) || ... || -works(eC,s) || allWorkIn(s, <e1..eC>)
	   
	    which is in CNF.
	*/
	
		if ($DEBUG) addComment("   c (1.1)");		
		/*(1.1) AND_{s in 1..7T} 
		 		OR_{ <e1...eC> (ordered) s.t. e1..eC in 1..I and all different } 
	   				allWorkIn(s, <e1..eC>)
		*/
		for($s=1; $s<=7*$T; $s++) {
			$clause = "";
			$tuple = firstOrderedTuple($C,1, $I);  // a C-tuple with elements in 1..I
			while ($tuple !== NULL) {
				$clause .= allWorkIn2dimacs($s,$tuple) . " ";
				$tuple = nextOrderedTuple($tuple, $C,1,$I);
			}
			addClause($clause);
		}
	
		if ($DEBUG) addComment("   c (1.2a)");	
		/* (1.2a) AND_{s in 1..7T,  <e1..eC> as above} 
	   			-allWorkIn(s, <e1..eC>) || works(e1,s) 
	   			AND
	   			...
	   			AND
	   			-allWorkIn(s, <e1..eC>) || works(eC,s) 
	   	*/
	   	for($s=1; $s<=7*$T; $s++) {
		   	$tuple = firstOrderedTuple($C,1, $I);  // a C-tuple with elements in 1..I
		   	while ($tuple !== NULL) {
		   		foreach($tuple as $elem) {
			   		addClause("-".allWorkIn2dimacs($s,$tuple) . " " . works2dimacs($elem,$s) );
			   	}
			   	$tuple = nextOrderedTuple($tuple, $C,1,$I);
		   	}
	   	}
	   	
		if ($DEBUG) addComment("   c (1.2b)");	
	    /* (1.2b) AND_{s in 1..7T,  <e1..eC> as above} 
   			-works(e1,s) || ... || -works(eC,s) || allWorkIn(s, <e1..eC>)
	   	*/
	   	for($s=1; $s<=7*$T; $s++) {
		   	$tuple = firstOrderedTuple($C,1, $I);  // a C-tuple with elements in 1..I
		   	while ($tuple !== NULL) {
		   		$clause = "";
		   		foreach($tuple as $elem) {
			   		$clause .= "-".works2dimacs($elem,$s). " ";
			   	}
			   	$clause .= allWorkIn2dimacs($s, $tuple);
			   	addClause($clause);
			   	$tuple = nextOrderedTuple($tuple, $C,1,$I);
		   	}
	   	}
	   	
	
	

	if ($DEBUG) addComment("c (2) Every employee, after working (not in the night), must not work for the next $R slots");
	/* 	AND_{e in I, s in 1..7*T s.t. !(1 <= s%T <= N) [i.e., in a slot not in the night, cf. below] }
			works(e,s) -> 
				AND_{ s' in s+1..s+$R [sum made modulo 7T] }   -works(e,s').
				
		In other words, if an employee `e' works in slot `s' (not in the night), she must be free in the following $R slots. Since the scheduling is periodic, we have to consider summation modulo 7T (night slots are 1..N (first day), T+1..T+N (second day), ..., kT+1..kT+N (k-th day),...,  hence those slots `s' such that 
		1<= s%T <= N).
		
		This constraint can be rewritten into:
			AND_{e in I, s in 1..7*T s.t. !(1 <= s%T <= N)}		
				-works(e,s)  ||  AND_{ s' in s+1..s+$R [sum made modulo 7T] } -works(e,s').
		
		and, by distributing OR over AND:
			AND_{e in I, s in 1..7*T s.t. !(1 <= s%T <= N)}				
				AND_{ s' in s+1..s+$R [sum made modulo 7T] } -works(e,s) || -works(e,s').		

		The latter form is in CNF.				
	*/
  	for($e=1; $e<=$I; $e++) {	
	   	for($s=1; $s<=7*$T; $s++) {		
  			if ( $s % $T < 1   ||  $s % $T > $N ) {
			   	for($ss=$s+1; $ss<=$s+$R; $ss++) {		  			
			   		$sss = $ss;
			   		if ($sss > 7*$T) $sss = $sss % 7*$T;
  					addClause("-" . works2dimacs($e,$s) . " -" . works2dimacs($e,$sss));
  				}
  			}
  		}
	}
	

	if ($DEBUG) addComment("c (3) Every employee, after working during night, must not work until the end of the next day");
	/* AND_{e in I, s in 1..7T s.t. (1 <= s%T <= N) [i.e., in a slot in the night] }
			works(e, s ) -> 
				AND_{ s' in s+1..s+k, with k>0 the min. integer s.t. (s+k)%T = 0 and k>=T [end of next day] [sum made modulo 7T] } -works(e,s')
				
				
	   This can be straightforwardly rewritten in CNF by replacing `works(e,s) -> AND_...' with `AND_...  works(e,s) -> ...':

			AND_{e in I, s in 1..7T s.t. (1 <= s%T <= N) [i.e., in a slot in the night] }
				AND_{ s' in s+1..s+k, with k>0 the min. integer s.t. (s+k)%T = 0 and k>=T [end of next day] [sum made modulo 7T] }
						-works(e, s ) || -works(e,s')
	*/
  	for($e=1; $e<=$I; $e++) {	
	   	for($s=1; $s<=7*$T; $s++) {		
  			if ( ($s % $T) >= 1   &&    ($s % $T) <= $N ) {
  				//addComment("AND_{ s' in $s+1..$s+k, with k>0 the min. integer s.t. ($s+k) % $T = 0 and k>=$T [end of next day] [sum made modulo 7*$T] } -works($e, $s ) || -works($e,s')");
  				$k=1;
				while ($k<=$T  || ( (($s+$k) % $T) != 1 ) ) {
						$ss = $s+$k;				
						if ($ss > 7*$T)  $ss = $ss - 7*$T;
						addClause( "-".works2dimacs($e,$s) . " -".works2dimacs($e,$ss) );
						$k++;
				} 
			}
		}
	}
	
	

	if ($DEBUG) addComment("c (4) Employee either don't work at all, or work for at least $L slots");
	/* AND_{e in I}
			( AND_{ s in 1..7T } -works(e,s) )
			OR
			OR_{ <s1,...sL> (ordered) s.t. s1..sL in 1..7*T and all different } 
					works(e, s1) & ... & works(e, sL)
					
		In other words, for any employee, either she does never work during week, or there exist
		at least $L distinct slots to which she is assigned (considering only ordered tuples
		avoids to generate duplicate clauses, as usual).
		
		Also this constraint is not in CNF, and must be converted.
		Again, we resort to additional variables:
		
		   neverWorks(e) = true iff employee `e' never works
		   worksIn(e, <s1..sL>) = true  iff  employee `e' works at least in slots s1..sL
		
		Constraint becomes:
		
		(4.1) AND_{e in I}
			  neverWorks(e)
			OR
			  OR_{ <s1,...sL> (ordered) s.t. s1..sL in 1..7*T and all different } 
					worksIn(e, <s1..sL>)
		
		AND
		
		(4.2) AND_{e in I}
				neverWorks(e) <-> ( AND_{ s in 1..7T } -works(e,s) )
		
		AND
		
		(4.3) AND_{e in I} AND_{ <s1,...sL> (ordered) s.t. s1..sL in 1..7*T and all different } 
				worksIn(e, <s1,..,sL>) <-> works(e, s1) & ... & works(e, sL)
		
		
		As in the previous constraint, (4.1) is in CNF (it is a set of long clauses, one for each employee), while (4.2) and (4.3) become (splitting `<->' as usual):
		
		(4.2a) AND_{e in I} AND_{ s in 1..7T } neverWorks(e) -> -works(e,s)
		AND
		(4.2b) AND_{e in I} ( OR_{ s in 1..7T } works(e,s) ) || neverWorks(e)
		
		AND
		
		(4.3a) AND_{e in I} AND_{ <s1,...sL> (ordered) s.t. s1..sL in 1..7*T and all different } 
				worksIn(e, <s1,..,sL>) -> works(e, s1) 
				& 
				... 
				& 
				worksIn(e, <s1,..,sL>) -> works(e, sL)
				
		 AND
		 
		 (4.3b) AND_{e in I} AND_{ <s1,...sL> (ordered) s.t. s1..sL in 1..7*T and all different } 
		 		-works(e, s1) || ... || -works(e, sL) || worksIn(e, <s1,..,sL>).
		
		 All such constraints are in CNF.
	*/


	if ($DEBUG) addComment("\n\nc (4.1)");
	/*	(4.1) AND_{e in I}
			  neverWorks(e)
			OR
			  OR_{ <s1,...sL> (ordered) s.t. s1..sL in 1..7*T and all different } 
					worksIn(e, <s1..sL>)
	*/
  	for($e=1; $e<=$I; $e++) {	
  		$clause = neverWorks2dimacs($e);
	   	$tuple = firstOrderedTuple($L,1, 7*$T);  // a L-tuple with elements in 1..7*T
	   	while ($tuple !== NULL) {
	   		$clause .= " ".worksIn2dimacs($e, $tuple);
			$tuple = nextOrderedTuple($tuple, $L,1, 7*$T);	   		
	   	}
		addClause($clause);
	}
	

	if ($DEBUG) addComment("\n\nc (4.2a)");
	/*(4.2a) AND_{e in I} AND_{ s in 1..7T } neverWorks(e) -> -works(e,s)
	*/
  	for($e=1; $e<=$I; $e++) {		
	   	for($s=1; $s<=7*$T; $s++) {		
	   		addClause("-".neverWorks2dimacs($e) . " -".works2dimacs($e,$s));
	   	}
	}
		
	if ($DEBUG) addComment("\n\nc (4.2b)");	
	/*	(4.2b) AND_{e in I} ( OR_{ s in 1..7T } works(e,s) ) || neverWorks(e)
	*/
  	for($e=1; $e<=$I; $e++) {		
  		$clause = "";
	   	for($s=1; $s<=7*$T; $s++) {
	   		$clause .= works2dimacs($e,$s) . " ";
	   	}
	   	$clause .= neverWorks2dimacs($e);
	   	addClause($clause);
	}


	if ($DEBUG) addComment("\n\nc (4.3a)");	
	/*	(4.3a) AND_{e in I} AND_{ <s1,...sL> (ordered) s.t. s1..sL in 1..7*T and all different } 
				worksIn(e, <s1,..,sL>) -> works(e, s1) 
				& 
				... 
				& 
				worksIn(e, <s1,..,sL>) -> works(e, sL)
	*/
  	for($e=1; $e<=$I; $e++) {
	   	$tuple = firstOrderedTuple($L,1, 7*$T);  // a L-tuple with elements in 1..7*T
	   	while ($tuple !== NULL) {
	   		for ($idx=0; $idx<count($tuple); $idx++) {
	   			$s = $tuple[$idx];
		   		addClause( "-".worksIn2dimacs($e, $tuple) . " " . works2dimacs($e, $s) );
		   	}
			$tuple = nextOrderedTuple($tuple, $L,1, 7*$T);	   				   				   	
		}
	}
  	
 	
	if ($DEBUG) addComment("\n\nc (4.3b)");		
	/*(4.3b) AND_{e in I} AND_{ <s1,...sL> (ordered) s.t. s1..sL in 1..7*T and all diff. } 
		 		-works(e, s1) || ... || -works(e, sL) || worksIn(e, <s1,..,sL>).
	*/
  	for($e=1; $e<=$I; $e++) {
	   	$tuple = firstOrderedTuple($L,1, 7*$T);  // a L-tuple with elements in 1..7*T
	   	while ($tuple !== NULL) {
	   		$clause = "";
	   		foreach($tuple as $s) {
	   			$clause .= "-".works2dimacs($e, $s). " ";
	   		}
	   		$clause .= worksIn2dimacs($e, $tuple);
	   		addClause($clause);
			$tuple = nextOrderedTuple($tuple, $L,1, 7*$T);	   				   			   		
	   	}
	}


	// Print out dictionary
	echoDictionary();
	
	// Print out a list of structural properties of variables (useful for ad-hoc solvers)
	echoProperties();
	
	
	// Print out header line "p cnf..."
	isset($NVARS) or die("maxVar is not set");
	echo("p cnf $NVARS $NCLAUSES\n");
	
	
	// Print out all clauses
	echoClauses();	

	//print_r($dictionary);
	
	


	function addClause($clause) {
		global $CLAUSES,$NCLAUSES;
		$clause = trim($clause);
		if ($clause == "") return;
		$clause .= " 0";
		$CLAUSES[] = $clause; // Append to array
		$NCLAUSES++;
	}

	function addComment($msg) {
		global $CLAUSES;
		$CLAUSES[] = $msg;
		// No $NCLAUSES increment here!
	}


	// 	dictionary['WORKS'][empl][slot] =  (works(empl,slot)==true)
	function works2dimacs($e,$s) {
		global $NVARS, $dictionary, $DEBUG, $I, $T;
		
		if ($e < 1 || $e > $I) {
			echoClauses();
			die("Trying to retrieve a 'WORKS' variable for employee $e: WORKS($e,$s)");
		}
		if ($s < 1 || $s > 7*$T) {
			echoClauses();		
			die("Trying to retrieve a 'WORKS' variable for slot $s: WORKS($e,$s)");
		}
		$result = "WORKS(e=$e,s=$s)";		
		
		$v = $dictionary['WORKS'][$e][$s];
		if ($v !== NULL) {
			if (!$DEBUG) $result = $v;
		}
		else {
			$NVARS++;
			$dictionary['WORKS'][$e][$s] = $NVARS;
			if (!$DEBUG) $result = $NVARS;
		}
		return $result;
	}
	
	
	//dictionary['ALL_WORK_IN][slot][tupleOfEmployees] = (allWorkIn(s, <e1...eC>)==true)
	function allWorkIn2dimacs($s,$tuple) {
		global $NVARS, $dictionary, $DEBUG, $I, $T;
		$tupleString = join(",",$tuple); // tuple of employees		
		if ($s < 1 || $s > 7*$T) {
			echoClauses();		
			die("Trying to retrieve a 'ALL_WORK_IN' variable for slot $s: ALL_WORK_IN($s,emp=<$tupleString>)");
		}

		$result = "ALL_WORK_IN(s=$s,empl=<".$tupleString.">)";
		
		$v = $dictionary['ALL_WORK_IN'][$s][$tupleString];
		if ($v != NULL) {
			if (!$DEBUG) $result = $v;
		}
		else {
			$NVARS++;
			$dictionary['ALL_WORK_IN'][$s][$tupleString] = $NVARS;
			if (!$DEBUG) $result = $NVARS;
		}
		return $result;
	}	
	
	
	// dictionary['WORKS_IN'][empl][tupleOfSlots] = (worksIn(e, <s1..sL>) = true)
	function worksIn2dimacs($e,$tuple) {
		global $NVARS, $dictionary, $DEBUG, $I, $T;
		
		if (!is_array($tuple)) {
			die("tuple is not array: " . ($tuple===null?"null":$tuple)."\n");
		}
		
		$tupleString = join(",", $tuple); // tuple of slots		
		if ($e < 1 || $e > $I) {
			echoClauses();		
			die("Trying to retrieve a 'WORKS_IN' variable for employee $e: WORKS_IN(emp=$e,slots=<$tupleString>)");
		}

		$result = "WORKS_IN(e=$e,slots=<".$tupleString.">)";
		
		$v = $dictionary['WORKS_IN'][$e][$tupleString];
		if ($v != NULL) {
			if (!$DEBUG) $result = $v;
		}
		else {
			$NVARS++;
			$dictionary['WORKS_IN'][$e][$tupleString] = $NVARS;
			if (!$DEBUG) $result = $NVARS;
		}
		return $result;
	}

	
	
	// dictionary['NEVER_WORKS][employee] = (neverWorks(empl) == true)
	function neverWorks2dimacs($e) {
		global $NVARS, $dictionary, $DEBUG, $I, $T;
		
		if ($e < 1 || $e > $I) {
			echoClauses();		
			die("Trying to retrieve a 'NEVER_WORKS' variable for employee $e: NEVER_WORKS($e)");
		}

		$result = "NEVER_WORKS(e=$e)";
		
		$v = $dictionary['NEVER_WORKS'][$e];
		if ($v != NULL) {
			if (!$DEBUG) $result = $v;
		}
		else {
			$NVARS++;
			$dictionary['NEVER_WORKS'][$e] = $NVARS;
			if (!$DEBUG) $result = $NVARS;
		}
		return $result;
	}	
	
	
	function echoClauses() {
		global $CLAUSES;
		foreach($CLAUSES as $c) {
			echo($c ."\n");
		}
	}	
	
	

	
	
	function echoDictionary() {
		global $NVARS, $dictionary, $DEBUG, $I, $T;
		echo ("c\nc START DICTIONARY\n");
		/*if ($DEBUG) {
			echo("c omitted -- DEBUG mode\n");
		}
		else {*/		
			if (isset($dictionary['WORKS'])) {			
				// 	dictionary['WORKS'][empl][slot] =  dimacsVar
				foreach($dictionary['WORKS'] as $e => $entryE) {
					foreach($entryE as $s => $dimacsVar) {
						echo("c works($e,$s) = $dimacsVar\n");
					}
				}
			}		
			if (isset($dictionary['WORKS_IN'])) {
				// dictionary['WORKS_IN'][empl][tupleOfSlots] = dimacsVar
				foreach($dictionary['WORKS_IN'] as $e => $entryE) {
					foreach($entryE as $tuple => $dimacsVar) {
						echo("c works_in($e, <$tuple>) = $dimacsVar\n");
					}
				}	
			}
			if (isset($dictionary['ALL_WORK_IN'])) {
				//dictionary['ALL_WORK_IN][slot][tupleOfEmployees] = dimacsVar			
				foreach($dictionary['ALL_WORK_IN'] as $s => $entryS) {
					foreach($entryS as $tuple => $dimacsVar) {
						echo("c all_work_in($s, <$tuple>) = $dimacsVar\n");
					}
				}	
			}
			if (isset($dictionary['NEVER_WORKS'])) {
				// dictionary['NEVER_WORKS][employee] = dimacsVar
				foreach($dictionary['NEVER_WORKS'] as $e => $dimacsVar) {
					echo("c never_works($e) = $dimacsVar\n");
				}	
			}			
		//}
		echo("c END DICTIONARY\nc\nc\n");		
	}	
	
	
	function echoProperties() {
		global $NVARS, $dictionary, $DEBUG, $I, $T;
		echo("c START PROPERTIES\n");
		if (isset($dictionary['WORKS_IN'])) {
			/* WORKS_IN variables are dependent on some WORK variables
				worksIn(e, <s1,..,sL>) <-> works(e, s1) & ... & works(e, sL)
		    */	
			foreach($dictionary['WORKS_IN'] as $e => $entryE) {
				// dictionary['WORKS_IN'][empl][tupleOfSlots] = dimacsVar
				foreach($entryE as $tuple => $dimacsVar) {
					$tupleArray = explode(",", $tuple);
					echo("c " . worksIn2dimacs($e, $tupleArray)." DEPENDENT ON ");
					foreach($tupleArray as $s) {
						echo( works2dimacs($e, $s) . " ");
					}
					echo("\n");
				}
			}
		}
				

		if (isset($dictionary['ALL_WORK_IN'])) {
			/* ALL_WORK_IN variables are dependent on some WORKS variables
	        	allWorkIn(s, <e1...eC>)  <-> works(e1,s) & ... & works(eC,s)
		    */	
			foreach($dictionary['ALL_WORK_IN'] as $s => $entryS) {
				//dictionary['ALL_WORK_IN][slot][tupleOfEmployees] = dimacsVar			
				foreach($entryS as $tuple => $dimacsVar) {
					$tupleArray = explode(",", $tuple);
					echo("c " . allWorkIn2dimacs($s, $tupleArray)." DEPENDENT ON ");
					foreach($tupleArray as $e) {
						echo( works2dimacs($e, $s) . " ");
					}
					echo("\n");
				}
			}
		}

		if (isset($dictionary['NEVER_WORKS'])) {
			/* NEVER_WORKS variables are dependent on some WORKS variables
	        	neverWorks(e) <-> ( AND_{ s in 1..7T } -works(e,s) )
		    */	
			foreach($dictionary['NEVER_WORKS'] as $e => $dimacsVar) {
				// dictionary['NEVER_WORKS][employee] = dimacsVar
				echo("c " . neverWorks2dimacs($e) . " DEPENDENT ON ");
				for($s=1; $s <= 7*$T; $s++) {
					echo( works2dimacs($e,$s) . " ");
				}
				echo("\n");
			}
		}
		echo("c END PROPERTIES\n");
	}	
	
	
	

	function firstOrderedTuple($len,$min, $max) {  // a $len-tuple with elements in 1..$max all different
		$result = array();
		for ($i = 0; $i<$len; $i++) {
			$result[$i] = $min+$i;
		}
		return $result;
	}

	function nextOrderedTuple($tuple, $len,$min, $max) {  // next $len-tuple (wrt $tuple) with elements in 1..$max all different
		$result = $tuple;

		for ($i=$len-1; $i>=0; $i--) {		
			$remaining = $len-1 - $i;
			if ($result[$i] < $max && ($result[$i]+1 <= $max-$remaining )   ) {
				$result[$i]++;			
				for ($j = $i+1; $j < $len; $j++) {
					$result[$j] = $result[$i] + ($j - $i);
				}			
				return $result;				
			}
		}
		return NULL;
	}
	
	
	
	
	function readInstanceFile($fileName) {
		global $I, $T, $N, $C, $R, $L;	
		$COMMENTS = array("#", "//");		
		$file = file($fileName) or die("Could not open instance file $fileName\n");
		foreach($file as $line) {
			//echo("c Reading line: $line\n");
			$commentSign = NULL;
			foreach($COMMENTS as $c) {
				 $pos = strpos($line, $c);
				 if ($pos === FALSE) continue;
				 if ($commentSign === NULL || $commentSign > $pos) $commentSign = $pos;
			}
			if ($commentSign !== NULL) {
				$line = substr($line, 0, $commentSign);
				//echo("   c Removing comments (from col. $commentSign onwards): $line\n");		
			}
			$line = trim($line);
			if ($line == "") continue;
	
			list($what,$content) = split("=",$line);
			$what = trim($what);
			$content = trim($content);
	
			//echo("   c What: $what -- Content: $content\n");		
			
			if ($what == "employees" && $I===NULL) {
				$I = $content;
			}
			else if ($what == "dailySlots" && $T===NULL) {
				$T = $content;
			}
			else if ($what == "nightSlots" && $N===NULL) {
				$N = $content;
			}
			else if ($what == "minEmployeesAtWork" && $C===NULL) {
				$C = $content;
			}
			else if ($what == "minBreak" && $R===NULL) {
				$R = $content;
			}
			else if ($what == "minWorkingSlots" && $L===NULL) {
				$L = $content;
			}
		}
	}	
?>
