c Uso: 
c - rinominare il file dandogli estensione .php, eliminando ".txt"
c - Eseguire: php shopScheduling.php <file istanza>
c
c Esecuzione con zChaff su istanza di test e successiva traduzione del modello trovato in forma intellegibile
c php shopScheduling.php ss-sampleInstance.txt > cnf.txt ; zchaff cnf.txt | php zchaff2ShopScheduling.php cnf.txt
c
c Shop scheduling to SAT compiler (the most straightforward encoding, just a simple exercise)
c Author: Daniele Zannella - Sapienza University of Rome
c Date: July 6, 2008
c
<?php

	$DEBUG = false;

	$COMMENTS = array("#", "//");

    $fileName = $argv[1];
	if ($fileName == "") {
		echo("c Syntax: php ".$argv[0]." <instance>\n");
		exit;
	};
	
	
	$CLAUSES = array();
	$OUTPUT_GATE = array();
	$NCLAUSES = 0;
	$NVARS = 0;
	$dictionary = NULL; // will be a 3D matrix: dictionary[$j][$m][$t] =  (start(j,m)==t)
	
	
?>
c
c Problem description (a variation of the Job Shop scheduling problem):
c We are given N jobs and M machines.
c Each job can be divided into M different tasks (durations of all tasks are given as input). 
c Task k has to be processed on machine k. It can start only after task k-1 (if k>1)
c is terminated on mackine k-1.
c
c The objective is to find a starting time for all tasks of all jobs such that 
c every job is terminates before a global deadline D.
c
c Input: N, M, D, T[j,k] forall j in 1..N and k in 1..M (duration of task k of job j).
c
<?php
	// Reading instance file
	$file = file($fileName) or die("Could not open instance file $fileName");
	
	// Variables
	$N = NULL;
	$M = NULL;
	$D = NULL;
	$T = NULL;  // will be a 2D-matrix:  T[job,machine] = duration of job j on machine m
	
	foreach($file as $line) {
		//echo("c Reading line: $line\n");
		$commentSign = NULL;
		foreach($COMMENTS as $c) {
			 $pos = strpos($line, $c);
			 if ($pos === FALSE) continue;
			 if ($commentSign === NULL || $commentSign > $pos) $commentSign = $pos;
		}
		if ($commentSign !== NULL) {
			$line = substr($line, 0, $commentSign);
			//echo("   c Removing comments (from col. $commentSign onwards): $line\n");		
		}
		$line = trim($line);
		if ($line == "") continue;

		list($what,$content) = split("=",$line);
		$what = trim($what);
		$content = trim($content);

		//echo("   c What: $what -- Content: $content\n");		
		
		if ($what == "jobs" && $N===NULL) {
			$N = $content;
			//echo("   c Setting nb of jobs: $N\n");					
		}
		else if ($what == "machines" && $M===NULL) {
			$M = $content;
			//echo("   c Setting nb of machines: $M\n");
		}
		else if ($what == "deadline" && $D===NULL) {
			$D = $content;		
			//echo("   c Setting deadline: $D\n");								
		}
		else {
			// Durations of job $what
			$durations = split(",",$content);
			$m = 1;
			foreach($durations as $d) {
				$T[$what][$m] = trim($d);
				//echo("   c Setting duration($what,$m): $d\n");									
				$m++;
			}
		}
	}
	
	// Check that everything has been given
	$N > 0 or die("Number of jobs missing\n");
	$M > 0 or die("Number of machines missing\n");
	$D > 0 or die("Deadline missing\n");
	
	for ($j=1; $j<=$N; $j++) {
		for ($m=1; $m<=$M; $m++) {
			if (!isset($T[$j][$m]) || !is_numeric($T[$j][$m]) || $T[$j][$m] <= 0) {
				die("Wrong duration for job $j on machine $m: " . $T[$j][$m]."\n");
			}
		}
	}
	// Everything is ok.
	
	echo("c Number of jobs: $N\n");
	echo("c Number of machines: $M\n");
	echo("c Deadline: $D\n");	
	
	echo("c Durations:\n");	
	for ($j=1; $j<=$N; $j++) {
		echo("c Job $j: ");
		for ($m=1; $m<=$M; $m++) {
			echo($T[$j][$m]." ");
		}
		echo("\n");
	}
?>
c
c 
c START DICTIONARY
<?php   
	if (!$DEBUG) {
		for ($j=1; $j<=$N; $j++) {
			for ($m=1; $m<=$M; $m++) {
				for ($t=1; $t<=$D; $t++) {		
					echo("c start($j,$m,$t) = " . var2dimacs($j,$m,$t)."\n");
				}
			}
		}
	} else echo("c omitted -- DEBUG mode\n");
?>
c END DICTIONARY
<?php
	
	
	//echo("@@@@@@@@ Primo vincolo \n");
	if ($DEBUG) addComment("c Each job has at least one starting time on each machine");
	// AND_{j in 1..N} AND_{m in 1..M}  OR_{t \in 1..D}  start[j,m,t];
	for ($j=1; $j<=$N; $j++) {
		for ($m=1; $m<=$M; $m++) {
			$NVARS++;
			array_push($OUTPUT_GATE, $NVARS);
			$clause = "6 -1 $NVARS ";
			for ($t=1; $t<=$D; $t++) {
				$clause .= var2dimacs($j,$m,$t) . " ";
			}
			addClause($clause);
		}
	}
	
	// queto vincoolo non in cnf posso risolverlo con l'implicazione materiale ovvero
	// forall(j to n) forall(m to M) forall(t1 to D) 'se' start[j,m,t1] -> (-start[j,m,t1+1] AND - start[j,m,t1+2] AND ... - start[j,m,t1+k]) dove 't1+k<=D'
	if ($DEBUG) addComment("c\nc\nc Each job has at most one starting time on each machine");
	//AND_{j in 1..N} AND_{m in 1..M} AND_{t1,t2 \in 1..D t.c. t1!=t2} 
	//			(!start[j,m,t1] || !start[j,m,t2]).
	// echo("c @@@@@@@@ Secondo vincolo \n");
	
/*	for ($j=1; $j<=$N; $j++) {
		for ($m=1; $m<=$M; $m++) {	
			for ($t1=1; $t1<=$D; $t1++) {
				$NVARS++;
				array_push($OUTPUT_GATE, $NVARS);
				$clause="10 -1 $NVARS ";
				$NVARS++;
				$auxOutput = $NVARS;
				$auxClause = "4 -1 $NVARS ";
				for($t2=$t1+1; $t2<=$D; $t2++) {
					$auxClause.= "-".var2dimacs($j,$m,$t2)." ";
				}
				addClause($auxClause);
				$clause.= var2dimacs($j,$m,$t1)." ".$auxOutput; 
				// adesso vado ad aggiungere la clausola finale alla lista 
				addClause($clause);
			}
		}
	}
	*/
	
	
	// il discorso fatto su è sintatticamente corretto, però SATMATE risolve "solo" istanze eDimacs con Gate AND, OR, NOT
	// dunque bisogna ristrutturare il tutto con queste porte	

	if ($DEBUG) addComment("c\nc\nc Each job has at most one starting time on each machine");
	//AND_{j in 1..N} AND_{m in 1..M} AND_{t1,t2 \in 1..D t.c. t1!=t2} 
	//			(!start[j,m,t1] || !start[j,m,t2]).
	for ($j=1; $j<=$N; $j++) {
		for ($m=1; $m<=$M; $m++) {	
			for ($t1=1; $t1<=$D; $t1++) {
				if($t1<$D) {
					//$NVARS++;
					//array_push($OUTPUT_GATE, $NVARS);
					//$clause="6 -1 $NVARS -".var2dimacs($j,$m,$t1)." ";
					if($t1<$D-1) {
						$NVARS++;
						$auxOutput = $NVARS;
						$auxClause = "4 -1 $auxOutput ";
						for($t2=$t1+1; $t2<=$D; $t2++) {
							$auxClause.= " -".var2dimacs($j,$m,$t2)." ";
							}
						addClause($auxClause);
						$NVARS++;
						array_push($OUTPUT_GATE, $NVARS);
						$clause="6 -1 $NVARS -".var2dimacs($j,$m,$t1). " ".$auxOutput."";
						addClause($clause);
					}
					else{
						$NVARS++;
						array_push($OUTPUT_GATE, $NVARS);
						$clause="6 -1 $NVARS -".var2dimacs($j,$m,$t1). " -".var2dimacs($j,$m,++$t1)."";
						addClause($clause);
						continue;
						}
				}	

			}
		}

	}





	// anche questo vincolo lo scriviamo medinate implicazione materiale
	if ($DEBUG) addComment("c\nc\nc Monotasking: forall j,m,t t.c. start(j,m)=t, there is no other j2 s.t. start(j2,m)=t', with t' in t..T[j,m]-1");	
	//AND_{j in 1..N} AND_{m in 1..M} AND_{t in 1..D} start[j,m,t] ->
	//	(AND_{j2 in 1..N, j2!=j} AND_{t2 in t..min(D,t+T[j,m]-1)}
	//		 !start[j2,m,t2]).
	//	echo("@@@@@@@@ Terzo vincolo \n");
	
	for ($j=1; $j<=$N; $j++) {
		for ($m=1; $m<=$M; $m++) {	
			for ($t=1; $t<=$D; $t++) {
					// ...
					if(($T[$j][$m]!=1)&&($D!=$t)) {
							$NVARS++;
							$auxClause1 ="";
							$auxOutput1 = $NVARS;
							$auxClause1 = "4 -1 $auxOutput1";
							for ($j2=1; $j2<=$N; $j2++) {
								if ($j2==$j) continue;
								for ($t2=$t; $t2<=min($D,$t+$T[$j][$m]-1); $t2++) {
									$auxClause1 .= " -" . var2dimacs($j2,$m,$t2);
									}
							}
							addClause($auxClause1);
							$NVARS++;
							array_push($OUTPUT_GATE, $NVARS);
							$clause = "6 -1 $NVARS -".var2dimacs($j,$m,$t)." ".$auxOutput1;
							addClause($clause);
						}
					else
						{
							for ($j2=1; $j2<=$N; $j2++) {
								if ($j2==$j) continue;
								$NVARS++;
								array_push($OUTPUT_GATE, $NVARS);
								$clause = "6 -1 $NVARS -".var2dimacs($j,$m,$t)." -" . var2dimacs($j2,$m,$t);
								addClause($clause);
							}			
						}
				}
		}
	}
	
	// ancora esprimiamo il tutto mediante implicazione materiale
	if ($DEBUG) addComment("c\nc\nc For each job j, processing on machine m>1 cannot start before processing on machine m-1 terminates");	
	//AND_{j in 1..N} AND_{m in 2..M} AND_{t in 1..D}
	//	start[j,m,t] -> (OR_{t2 in 1..t-T[j,m-1]} start[j,m-1,t2]).
	//	echo("@@@@@@@@ Quarto vincolo \n");

	for ($j=1; $j<=$N; $j++) {
		for ($m=2; $m<=$M; $m++) {	
			for ($t=1; $t<=$D; $t++) {
					if($t-$T[$j][$m-1] > 0) {
						// ...
						$NVARS++;
						$auxOutput2 = $NVARS;
						
						if($t-$T[$j][$m-1] == 1)
						{
							array_push($OUTPUT_GATE, $NVARS);
							$clause = "6 -1 $auxOutput2 -".var2dimacs($j,$m,$t)." ".var2dimacs($j,($m-1), 1);
							addClause($clause);
							continue;
						}
						else {
							$auxClause = "";
							for ($t2=1; $t2<=$t-$T[$j][$m-1]; $t2++) {
								$auxClause .= var2dimacs($j,$m-1,$t2) . " ";
							}
							//addClause($auxClause2);
							$NVARS++;
							array_push($OUTPUT_GATE, $NVARS);
							$clause = "6 -1 $NVARS -".var2dimacs($j,$m,$t)." ".$auxClause;
							//$clause .= $auxOutput2;
							addClause($clause);
						}
					}
					else
					{
						$NVARS++;
						array_push($OUTPUT_GATE, $NVARS);
						// NOT...
						$clause = "3 -1 $NVARS ". var2dimacs($j,$m,$t);
						addClause($clause);
					}
			}
		}
	}

	
	$gateDead = array();
	if ($DEBUG) addComment("c\nc\nc Deadline respected");	
	//AND_{j in 1..N} AND_{m in 1..M} (OR_{t in 1..D-T[j,m]} start[j,m,t]).
	//echo("@@@@@@@@ Quinto vincolo \n");
	
	for ($j=1; $j<=$N; $j++) {
		for ($m=1; $m<=$M; $m++) {
			$var = $D-$T[$j][$m];
			if($var>1) {
					$NVARS++;
					array_push($OUTPUT_GATE, $NVARS);	
					$clause = "6 -1 $NVARS ";
					for ($t=1; $t<=$D-$T[$j][$m]; $t++) {
							$clause .= var2dimacs($j,$m,$t) . " ";
						}
					addClause($clause);
				}

			else if($var==1)
			{
				$clause1 ="";
				$NVARS++;
				array_push($OUTPUT_GATE, $NVARS);	
				$clause1 .="3 -1 $NVARS -" .var2dimacs($j,$m,1);	
				addClause($clause1);			
			}

		}
	}
	
	
	//Terminati di scrivere tutti i vincoli singolarmente li vado a mettere tutti in AND per avere un unico 'Gate' (o root) di uscita finale. 
	$NVARS++;
	$finalClause = "4 -1 $NVARS ";
	for($k=0;$k<=count($OUTPUT_GATE);$k++)
		{
			$finalClause .= $OUTPUT_GATE[$k]." "; 
		}
	addClause($finalClause);
	
	// Print out header line "p Non-Cnf..."
	isset($NVARS) or die("maxVar is not set");
	echo("p noncnf $NVARS\n");
	
	// Print out all clauses
	echoClauses();	

	
	
	


	function addClause($clause) {
		global $CLAUSES,$NCLAUSES;
		$clause = trim($clause);
		if ($clause == "") return;
		$clause .= " 0";
		$CLAUSES[] = $clause; // Append to array
//		$NCLAUSES++;
	}

	function addComment($msg) {
		global $CLAUSES;
		$CLAUSES[] = $msg;
		// No $NCLAUSES increment here!
	}


	function var2dimacs($j,$m,$t) {
		global $NVARS, $dictionary, $DEBUG, $N, $M, $D;
		
		if ($j < 1 || $j > $N) {
			echoClauses();
			die("Trying to retrieve a variable for job $j: start($j,$m,$t)");
		}
		if ($m < 1 || $m > $M) {
			echoClauses();		
			die("Trying to retrieve a variable for machine $m: start($j,$m,$t)");
		}
		if ($t < 1 || $t > $D) {
			echoClauses();		
			die("Trying to retrieve a variable for time $t: start($j,$m,$t)");
		}


		
		if ($DEBUG) return "start(j=$j,m=$m,t=$t)";
		
		$v = $dictionary[$j][$m][$t];
		if ($v !== NULL) return $v;
		else {
			$NVARS++;
			$dictionary[$j][$m][$t] = $NVARS;
			return $NVARS;
		}
	}
	
	
	function echoClauses() {
		global $CLAUSES;
		foreach($CLAUSES as $c) {
			echo($c ."\n");
		}
	}		
?>
